<?php
if (!defined('MEDIAWIKI'))
	die("MediaWiki extensions cannot be run directly.");
/**
 * WikiDB
 * A MediaWiki extension which adds database functionality to your wiki, which
 * behaves in a truly wiki-like way.
 *
 * The extension is fairly well developed, but there are a lot more features
 * in the pipeline.  Visit my test wiki for documentation and updates, to get help
 * or to leave feedback.  See $pWikiDB_URL (below) for the URL.  Bug reports,
 * patches and feature suggestions welcome.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2017, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 1365 $
 */

// WikiDB requires MW 1.7 or above.
	if (version_compare($wgVersion, '1.7', '<')) {
		WikiDB_FatalError("Incompatible MediaWiki Extension",
						  "WikIDB cannot be installed on MediaWiki versions earlier "
						  . "than 1.7.0.  You need to disable the extension, or "
						  . "upgrade your copy of MediaWiki.");
	}

//////////////////////////
// Credit information

// The version is an integer, and is incremented manually every time changes are
// made that require command-line update scripts to be run.  This includes both
// changes to the DB table schema and changes that might invalidate existing WikiDB
// data, requiring a rebuild.
// Changes that will not break existing installations of WikiDB do not cause
// this number to be updated.
	$pWikiDB_Version = 4;

// The revision is also an integer, and will hold the highest revision number of
// any WikiDB files.  Each file needs to call WikiDB_RegisterRevision() with its
// SVN revision string, and the highest number registered in this way will be
// used as the revision number for the extension.
// This is held in $pWikiDB_Revision, which is initialised to 0.
// This does not need any manual intervention so long as I remember to add a call to
// WikiDB_RegisterRevision() at the top of any new files, and set the svn:keywords
// property correctly.
// Note that changes to files in the maintenance directory do not cause the
// revision number to update.  Not sure if I can reasonably do anything about that!
	$pWikiDB_Revision = 0;
	WikiDB_RegisterRevision('$Rev: 1365 $');

// Set these variables here so they can be re-used for each registered item.
// Needs to be done before SpecialPages.php is included, as they are used there
// too.
	$pWikiDB_Author = "Mark Clements";
	$pWikiDB_URL = "http://www.kennel17.co.uk/testwiki/WikiDB";


//////////////////////////
// Include files

// Need to ensure that the IncludableSpecialPage class is defined before we define
// our own special pages, so include the appropriate MW class file (if not already
// included).
// This was moved in MW 1.23, so we need to check whether it is in the old location
// or the new one, before loading it.
	if (file_exists($IP . "/includes/SpecialPage.php"))
		require_once($IP . "/includes/SpecialPage.php");
	else
		require_once($IP . "/includes/specialpage/SpecialPage.php");

// Include all the other WikiDB files:
// Need to define WikiDB to let them know that it is safe to load.
	define("WikiDB", true);

// We calculate the include path here, so it only needs to be done once.
	$pWikiDB_IncludePath = dirname(__FILE__) . "/includes/";

// Support files
	require_once($pWikiDB_IncludePath . "Constants.php");
	require_once($pWikiDB_IncludePath . "Utility.php");
	require_once($pWikiDB_IncludePath . "Regex.php");
	require_once($pWikiDB_IncludePath . "BuiltInTypes.php");

// Core WikiDB classes
	require_once($pWikiDB_IncludePath . "classWikiDB.php");
	require_once($pWikiDB_IncludePath . "classWikiDB_Table.php");
	require_once($pWikiDB_IncludePath . "classWikiDB_Query.php");
	require_once($pWikiDB_IncludePath . "classWikiDB_QueryResult.php");
	require_once($pWikiDB_IncludePath . "classWikiDB_Parser.php");
	require_once($pWikiDB_IncludePath . "classWikiDB_HTMLRenderer.php");
	require_once($pWikiDB_IncludePath . "classWikiDB_TypeHandler.php");

// MW interface hooks
	require_once($pWikiDB_IncludePath . "Hooks.php");
	require_once($pWikiDB_IncludePath . "ParserTags.php");
	require_once($pWikiDB_IncludePath . "SpecialPages.php");
//	require_once($pWikiDB_IncludePath . "MagicWords.php");

// MW 1.19 changed the way that Wiki::getAction() function was implemented, so it
// now only returns actions for which there is an appropriate Action class defined.
// Previously, it would return whatever was passed in via the URL.  Therefore we
// need to define an action class in order to work correctly in MW 1.19 and above.
// From my testing, the only issue if we omit this is that the 'viewdata' tab is
// not correctly selected when viewing data (the actual page itself renders fine).
// However, there may be other issues that I've not spotted.
// Therefore, for MW 1.19 and above, we need to initialise our viewdata action via
// a special class.
	if (version_compare($wgVersion, '1.19', '>=')) {
		require_once($pWikiDB_IncludePath . "Actions.php");
		$wgActions['viewdata'] = "WikiDB_ViewDataAction";
	}

	unset($pWikiDB_IncludePath);

//////////////////////////
// Extension credits:
// Note that special page credits are registered in includes/SpecialPages.php.

// Main extension credits
// Note that 'description' is used as the extension description in MW < 1.12.
// From 1.12 onwards, 'descriptionmsg' is used to load the specified translatable
// string from the message store.  If either of these is changed, remember to ensure
// that the other is also updated to match!
	$wgExtensionCredits['other'][] = array(
		'path' => __FILE__,
		'name' => "WikiDB",
		'author' => $pWikiDB_Author,
		'description' => "a wiki-based database system",
		'descriptionmsg' => 'wikidb-description',
		'version' => $pWikiDB_Version . ", r" . $pWikiDB_Revision,
		'url' => $pWikiDB_URL,
		'license-name' => "CC-BY-SA-2.0 (UK)",
	);


//////////////////////////
// Variable initialisation

// If $wgWikiDBNamespaces has not been set, initialise it to an empty array.
	if (!isset($wgWikiDBNamespaces))
		$wgWikiDBNamespaces = array();

// $wgWikiDBDefaultClasses can be used to specify a class string to be used for all
// data tables output by WikiDB.  If this variable has not been set, it defaults
// to MediaWiki's built-in "wikitable" class.  Note that the "WikiDBDataTable" class
// will always be added to this class string before use so that there is a consistent
// class name in the list, e.g. to make it accessible to JS code.
	if (!isset($wgWikiDBDefaultClasses))
		$wgWikiDBDefaultClasses = "wikitable";

// If $wgWikiDBMaxRefreshRate has not been set, initialise it to the default value.
// This variable can be either set to a number indicating the number of
// stale rows to refresh on each page view, or to one of the following constants:
// 	WIKIDB_DisableAutoRefresh
//		Disables this auto-refresh completely (presumably because you are handling
//		it via a cron job).
// 	WIKIDB_RefreshAll
//		Update all stale rows on every page request (if there are any), so there
//		is never any stale data.
// The optimum value for this setting will depend on how many visitors you get,
// how large your tables are (in terms of numbers of rows) and how often your table
// definitions change.  The default has been chosen as a good average-case
// value, but you will probably want to customise this.
// Note that the recommended method of refreshing stale links is to run
// RefreshStaleData.php via a regular cron job, in which case you should set this
// variable to WIKIDB_DisableAutoRefresh.
	if (IsBlank(@$wgWikiDBMaxRefreshRate))
		$wgWikiDBMaxRefreshRate = pWIKIDB_DefaultMaxRefreshRate;

//////////////////////////
// Extension registration

	$wgExtensionFunctions[] = "efWikiDB";

// It seems that in MW < 1.18, we can't load our message file inside our startup
// function - perhaps it is too late in the process?  Therefore we set it up here.
	global $pWikiDB_MessagesFile;
	$pWikiDB_MessagesFile = dirname( __FILE__ ) . '/WikiDB.i18n.php';

	if (isset($wgExtensionMessagesFiles))
		$wgExtensionMessagesFiles['WikiDB'] = $pWikiDB_MessagesFile;

	function efWikiDB() {
		global $wgMessageCache, $wgHooks, $wgParser, $wgVersion, $wgCommandLineMode;
		global $pWikiDB_Messages, $wgWikiDBMaxRefreshRate, $pWikiDB_MessagesFile;
		global $wgExtensionMessagesFiles;

	// Localisable messages
		if (!isset($wgExtensionMessagesFiles)) {
			require_once($pWikiDB_MessagesFile);
			$wgMessageCache->addMessages($messages['en']);
		}

	// Application hooks
		$wgHooks['UnknownAction'][] = "efWikiDB_SpecialActions";
		$wgHooks['ParserBeforeStrip'][] = "efWikiDB_ParserBeforeStrip";
		$wgHooks['ArticleDeleteComplete'][] = "efWikiDB_ArticleDeleteComplete";
		$wgHooks['TitleMoveComplete'][] = 'efWikiDB_TitleMoveComplete';

	// Although this hook has existed since MW 1.7, the necessary placeholder
	// properties that allow you to inject HTML into the form were only added in
	// MW 1.9.  As the hook is used solely to inject HTML in this manner (and as
	// there is no workaround for earlier versions) we only bother setting up this
	// hook on MW 1.9 and above.
		if (version_compare($wgVersion, '1.9', '>=')) {
			$wgHooks['EditPage::showEditForm:initial'][]
										= 'efWikiDB_EditPage_showEditForm_initial';
		}

	// Add hook to allow us to add our custom CSS files.
	// In early versions of MediaWiki, this was handled via the
	// SkinTemplateSetupPageCss hook, but this was removed in MW 1.19, with the
	// correct way now being to add the CSS in the BeforePageDisplay hook.
	// This hook was introduced in MW 1.7, but it wasn't until some later version
	// that the necessary addInlineStyle() method was added to the OutputPage class
	// (it is present in 1.17, but not sure when it was introduced).
	// We use the old method if that method is not available, or the new method if
	// it is.
		if (method_exists("OutputPage", "addInlineStyle"))
			$wgHooks['BeforePageDisplay'][] = "efWikiDB_AddCSS";
		else
			$wgHooks['SkinTemplateSetupPageCss'][] = "efWikiDB_AddCSS";

	// In MW 1.12 we now have a LinksUpdate hook which is run whenever the internal
	// links table is updated for a page.  This happens when a page is saved.
	// However it also happens when a template that the page uses is changed (which
	// is triggered via the job queue).  By using this hook we ensure that our
	// internal data structures are updated whenever the page content changes even
	// if the actual article itself is not directly changed.
	// For earlier versions, we still use the ArticleSaveComplete hook which works
	// for everything except template changes.  In these versions you need to make
	// a null edit to any affected pages whenever a template is modified.
	// Note that ArticleSaveComplete was deprecated in MW 1.28.
	// TODO: Are there other hooks from the above list that are no longer required
	//		 when we use LinksUpdate?  For example, do deletions/moves also trigger
	//		 this hook?
		if (version_compare($wgVersion, '1.12', '>=')) {
			$wgHooks['LinksUpdate'][] = 'efWikiDB_LinksUpdate';
		}
		else {
			$wgHooks['ArticleSaveComplete'][] = "efWikiDB_ArticleSaveComplete";
		}

	// MW 1.18 replaced some of the SkinTemplate* hooks with new hooks that have
	// been generalised, and which are designed around the new Vector skin.
	// We therefore need to set up an additional callback in order to ensure that
	// the 'viewdata' tab is present in 1.18.  In earlier versions, the
	// 'SkinTemplateContentActions' will be used.  In 1.18 and above, this hook no
	// longer exists, and 'SkinTemplateNavigation::Universal' is used instead.
	// As the old hook was removed at the same time as the new one was added, we
	// don't need to do a version check to avoid the hook being called twice.
		$wgHooks['SkinTemplateContentActions'][] = "efWikiDB_AddActionTabs";
		$wgHooks['SkinTemplateNavigation::Universal'][] = "efWikiDB_SkinHooks";

	// If MW version < 1.9.1, the ArticleUndelete hook wasn't present, so we need to
	// use a bit of a cheeky workaround...
		if (version_compare($wgVersion, '1.9.1', '<')) {
			$wgHooks['SpecialPageExecuteAfterPage'][]
								= 'efWikiDB_SpecialPageExecuteAfterPage';
		}
	// Otherwise we use ArticleUndelete
		else {
			$wgHooks['ArticleUndelete'][] = 'efWikiDB_ArticleUndelete';
		}

	// Parser hooks
	     $wgParser->setHook("data", "efWikiDB_DataTag");
	     $wgParser->setHook("repeat", "efWikiDB_RepeatTag");
	     $wgParser->setHook("guesstypes", "efWikiDB_GuessTypes");

	// On each page request, we update any stale data, up to the limit imposed
	// by $wgWikiDBMaxRefreshRate.  If that variable is set to the constant
	// WIKIDB_DisableAutoRefresh then auto-refresh is skipped (presumably because
	// a regular cron-job is handling it instead).
	// We also skip this in command-line mode, as there is a dedicated script for
	// this and we don't want to interfere with the running of that or any other
	// scripts.
	// In the most common case, where there is no stale data, the speed implication
	// should be negligible - if that is not the case then this will need revisiting.
		if (!$wgCommandLineMode
			&& $wgWikiDBMaxRefreshRate !== WIKIDB_DisableAutoRefresh)
		{
			WikiDB::RefreshStaleFieldData();
		}
	}


//////////////////////////
// Other registration functions

	function WikiDB_RegisterRevision($RevString) {
		global $pWikiDB_Revision;

		if (preg_match('/\$.*:\s?(\d+)\s?\$/', $RevString, $Matches)) {
			$Revision = intval($Matches[1]);
			if ($Revision > $pWikiDB_Revision)
				$pWikiDB_Revision = $Revision;
		}
	}

// WikiDB_FatalError()
// Local alternative to wfHttpError() as this isn't loaded at the point we want to
// use it, and moving the version check to later in the code means that MW may
// attempt to load PHP5 code on earlier PHP4-only versions.  To avoid this would
// mean a major architecture change (e.g. switching to the auto-loader, which would
// break our version numbering) so it's not something I plan to do at this point.
// Differences to wfHttpError()
// * $code is hard-coded to 500, so it doesn't need to be passed in.
// * We don't use $wgOut at all, as this isn't defined yet.
// * We die() at the end.
	function WikiDB_FatalError($label, $desc) {
		header( "HTTP/1.0 500 $label" );
		header( "Status: 500 $label" );

		header( 'Content-type: text/html; charset=utf-8' );
		print "<html><head><title>" .
			htmlspecialchars( $label ) .
			"</title></head><body><h1>" .
			htmlspecialchars( $label ) .
			"</h1><p>" .
			htmlspecialchars( $desc ) .
			"</p></body></html>\n";
		die();
	}
