<?php
if (!defined('MEDIAWIKI'))
	die("MediaWiki extensions cannot be run directly.");
if (!defined('WikiDB')) {
	die("The WikiDB extension has not been set up correctly.<br>'"
		. basename(__FILE__) . "' should not be included directly.<br>Instead, "
		. "add the following to LocalSettings.php: "
		. "<tt><b>include_once(\"WikiDB/WikiDB.php\");</b></tt>");
}
/**
 * Special pages provided by WikiDB.
 * The code layout for each IncludableSpecialPage object was originally copied from
 * the NewPages extension by Rob Church, with appropriate modifications.
 *
 * @package MediaWiki
 * @subpackage Extensions
 *
 * @author Mark Clements <mclements at kennel17 dot co dot uk>
 * @copyright Copyright  2006-2017, Mark Clements
 * @license http://creativecommons.org/licenses/by-sa/2.5/ cc-by-sa 2.5 or later
 * @version $Rev: 1370 $
 */

//////////////////////////////////////////////////////////////////////////////////
// !! IMPORTANT NOTE !!
//
// Correct JOIN handling was only added in MW 1.13.  From 1.10 to 1.12.x you
// could pre-build your JOINs and pass them in as the first argument (the
// 'tables' parameter), so long as you prefixed them with a space (if there is
// no space, the whole string will be quoted as an identifier).  Prior to 1.10
// this could not be acheived with the select() function - you would have to
// build the whole query yourself and use query() directly.
//
// TODO: Currently MW < 1.13 will give incorrect results on these pages, though
//		 I believe the lists are otherwise correct.  To what extent are we able
//		 to fix this?  It is probably too much effort to be worthwhile for these old
//		 MW versions, particularly as I am likely to drop support for them fairly
//		 soon.
// TODO: An alternative solution is to disable the special pages for versions on
//		 which they are known not to work.
//
// Note that this only affects these special pages, as far as I can see.  All
// other direct DB queries either use a signature compatible with both old and
// new MW versions, or use the query() function directly.
//////////////////////////////////////////////////////////////////////////////////

WikiDB_RegisterRevision('$Rev: 1370 $');

//////////////////////////
// Register special pages

	$wgSpecialPages['AllTables'] = 'WikiDB_SP_AllTables';
	$wgSpecialPages['UndefinedTables'] = 'WikiDB_SP_UndefinedTables';
	$wgSpecialPages['EmptyTables'] = 'WikiDB_SP_EmptyTables';

//////////////////////////
// WikiDB_SpecialPage
// This class extends the standard IncludableSpecialPage to add a couple of standard
// functions that WikiDB needs.

abstract class WikiDB_SpecialPage extends IncludableSpecialPage {

	var $pItemsPerPage = 50;

	function pRegister($SpecialPageName) {
	// Old PHP4-style way of calling the constructor - works in MW <= 1.16
		if (method_exists("SpecialPage", "SpecialPage")) {
			SpecialPage::SpecialPage($SpecialPageName, '', true, false,
									 'default', true);
		}
	// New PHP5-style - required for MW >= 1.17
		else {
			parent::__construct($SpecialPageName, '', true, false,
								'default', true);
		}
	}

	function execute($PathInfo) {
		global $wgOut;

		$fname = get_class($this) . "::" . __FUNCTION__;

		$this->pSetupPagination($PathInfo);
		$TextKeyPrefix = $this->GetTextKeyPrefix();

		$DB = wfGetDB(DB_SLAVE);

		$QueryInfo = $this->GetQuery($DB);
		$QueryInfo = $this->pTidyQueryInfo($QueryInfo);

	// Don't show the navigation if we're including the page.
		if (!$this->mIncluding) {
			$this->setHeaders();
			$wgOut->addWikiText(WikiDB_Msg($TextKeyPrefix . '-header'));

			$RowCount = $this->pGetCount($QueryInfo, $DB);
			if ($RowCount > 0)
				$this->pOutputPagination($RowCount);
		}

		if (count($QueryInfo['order']) > 0) {
			$QueryInfo['options']['ORDER BY'] = $QueryInfo['order'];
		}

		$QueryInfo['options']['LIMIT'] = $this->pItemsPerPage;
		$QueryInfo['options']['OFFSET'] = $this->pOffset;

		$res = $DB->select($QueryInfo['tables'], $QueryInfo['fields'],
						   $QueryInfo['conds'], $fname, $QueryInfo['options'],
						   $QueryInfo['join_conds']);
	// Ensure we return a proper result object on all versions of MediaWiki.
		$res = WikiDB_WrapDBResult($DB, $res);

		$RowCount = $res->numRows();
		if ($RowCount > 0) {
		// Make list
			$wgOut->addHTML("<ol>\n");
			while ($row = $res->fetchObject())
				$wgOut->addHTML($this->pMakeListItem($row));
			$wgOut->addHTML("</ol>\n");
		}
		else
			$wgOut->addWikiText(WikiDB_Msg($TextKeyPrefix . '-none'));
		$res->free();
	}

	abstract function GetQuery(&$DB);
	abstract function GetTextKeyPrefix();
	abstract function pMakeListItem($Row);

	function pSetupPagination($PathInfo) {
		global $wgRequest;

	// Get the 'items per page' from the path info provided.
		$ItemsPerPage = intval($PathInfo);

	// If the path info was set to zero, contained an invalid integer or was not
	// present at all, ignore it and instead use the 'limit' argument (falling-back
	// to our built-in default if this is also not set).
		if ($ItemsPerPage == 0) {
			$ItemsPerPage = $wgRequest->getIntOrNull('limit');
			if (!$ItemsPerPage)
				$ItemsPerPage = 50;
		}

		$this->pItemsPerPage = $ItemsPerPage;

	// Setup offset of current page (zero-based).
		$this->pOffset = $wgRequest->getInt('offset');
	}

	function pOutputPagination($TotalRows) {
		global $wgOut;

		$wgOut->addHTML('<p>'
						. wfShowingResults($this->pOffset, $this->pItemsPerPage)
						. '</p>');

		$LastRowOnPage = $this->pOffset + $this->pItemsPerPage;
		if ($LastRowOnPage >= $TotalRows)
			$IsLastPage = true;
		else
			$IsLastPage = false;

		$Output = WikiDB_GetPaginationLinks(
								 $this->pOffset,
								 $this->pItemsPerPage,
								 WikiDB_GetLocalSpecialPageName($this->mName),
								 "",
								 $IsLastPage
								);

		$wgOut->addHTML( '<p>' . $Output . '</p>' );
	}

	function pTidyQueryInfo($QueryInfo) {
		global $wgVersion;

		$arrExpectedFields = array('tables', 'fields', 'conds', 'order', 'options',
								   'join_conds');

		foreach ($arrExpectedFields as $Key) {
			if (isset($QueryInfo[$Key]))
				$QueryInfo[$Key] = (array) $QueryInfo[$Key];
			else
				$QueryInfo[$Key] = array();
		}

	// MW 1.18 and above support arguments being specified as arrays in all cases
	// where this is appropriate.  However, for older versions, we need to manually
	// collapse them into strings where necessary.
		if (version_compare($wgVersion, 1.18, "<")) {
			if (isset($QueryInfo['order']) && is_array($QueryInfo['order'])) {
				$QueryInfo['order'] = implode(", ", $QueryInfo['order']);
			}

			if (isset($QueryInfo['options']['GROUP BY'])
				&& is_array($QueryInfo['options']['GROUP BY']))
			{
				$QueryInfo['options']['GROUP BY']
								= implode(", ", $QueryInfo['options']['GROUP BY']);
			}
		}

	// Return the tidied array, with all expected elements present.
		return $QueryInfo;
	}

	function pGetCount($QueryInfo, $DB) {
		$fname = get_class($this) . "::" . __FUNCTION__;

		$DBResult = $DB->select($QueryInfo['tables'], $QueryInfo['fields'],
								$QueryInfo['conds'], $fname, $QueryInfo['options'],
								$QueryInfo['join_conds']);

	// Ensure we return a proper result object on all versions of MediaWiki.
		$DBResult = WikiDB_WrapDBResult($DB, $DBResult);

		$Count = 0;
		if ($DBResult) {
			$Count = $DBResult->numRows($DBResult);
			$DBResult->free();
		}

		return $Count;
	}

// pGetSkin()
// As of MW 1.18, the recommended method of getting the skin from within a special
// page is to call $this->getSkin().  The old method of pulling it from $wgUser
// is deprecated, and was removed in MW 1.27.
// To maintain backwards-compatibility with pre-1.18 versions, we have this wrapper
// function, which calls $this->getSkin() if it exists, otherwise uses the old
// $wgUser method.
	function pGetSkin() {
		global $wgUser;

		if (method_exists($this, "getSkin"))
			return $this->getSkin();
		else
			return $wgUser->getSkin();
	}

// pGetLinkToTitle()
// The method of generating a link from a Skin object (which extends the Linker
// class) has also changed.
// See WikiDB_MakeLink() for a description of what the changes were and when they
// occurred.
// TODO: Can this be reworked to use WikiDB_MakeLink(), to avoid the duplication?
//		 The behaviour is a bit different, as the SpecialPage class has access to
//		 some local context that isn't otherwise available, so it may not be
//		 possible.
	function pGetLinkToTitle($Title) {
		$Skin = $this->pGetSkin();

	// If this is MW >= 1.28, then the special page now has a LinkRenderer object
	// available to it, so we use that.
		if (method_exists($this, "getLinkRenderer")) {
			$LinkRenderer = $this->getLinkRenderer();
			return $LinkRenderer->makeLink($Title);
		}
	// If the $Skin object has a 'link' function, we call it.  This is required
	// for MW versions between the function being implemented and before it being
	// rewritten to work statically (where we need to call it on an object instance).
		elseif (method_exists($Skin, "link")) {
			return $Skin->link($Title);
		}
	// Alternatively, this might be an older MW version, from before the function
	// was introduced.  In this case, makeLinkObj() should be used.
		elseif (method_exists($Skin, "makeLinkObj")) {
			return $Skin->makeLinkObj($Title);
		}
	// If none of the above apply, we are on a MW version between the point where
	// Linker::link() became a static function and where the LinkRenderer class was
	// introduced (in MW 1.28).  In this case, the static function is the one to use.
		else {
			return Linker::link($Title);
		}
	}

} // END class WikiDB_SpecialPage


//////////////////////////
// Undefined Tables

class WikiDB_SP_UndefinedTables extends WikiDB_SpecialPage {

	function __construct() {
		parent::pRegister('UndefinedTables');
	}

	function GetTextKeyPrefix() {
		return "undefinedtables";
	}

	function GetQuery(&$DB) {
		$tblTables = $DB->tableName(pWIKIDB_tblTables);
		$tblRows = $DB->tableName(pWIKIDB_tblRows);

		$GroupFields = array(
					$tblRows . ".table_namespace",
					$tblRows . ".table_title",
				);

	// Later versions of MW allow us to use 'Alias' => "FieldDef", but to support
	// older versions we need to use "FieldDef AS Alias".
		$Fields = array(
					"COUNT(" . $tblRows . ".table_title) AS record_count",
				);
		$Fields = array_merge($Fields, $GroupFields);

		return array(
			'tables' => array(pWIKIDB_tblRows, pWIKIDB_tblTables),
			'fields' => $Fields,
			'conds' => array($tblTables . '.table_title IS NULL'),
			'join_conds' => array(
							pWIKIDB_tblTables => array(
									'LEFT JOIN',
									array(
											$tblRows . ".table_namespace = "
												. $tblTables . ".table_namespace",
											$tblRows . ".table_title = "
												. $tblTables . ".table_title",
										),
								),
						),
			'order' => array(
							"record_count DESC",
							$tblRows . ".table_namespace",
							$tblRows . ".table_title",
						),
			'options' => array(
							'GROUP BY'	=> $GroupFields,
						),
		);
	}

	function pMakeListItem($Row) {
		$Title = Title::makeTitleSafe($Row->table_namespace, $Row->table_title);
		if (!is_null($Title)) {
			$DefLink = $this->pGetLinkToTitle($Title);
			$DataLink = $Title->getLocalURL('action=viewdata');
			$DataLinkText = WikiDB_Msg('undefinedtables-record', $Row->record_count);

			return("<li>" . $DefLink . ' - <a href="' . $DataLink . '">'
				   . $DataLinkText . '</a></li>' . "\n");
		}
		else {
			return("<!-- Invalid title " . htmlspecialchars($Row->table_title)
				 . " in namespace " . htmlspecialchars($Row->table_namespace)
				 . " -->\n");
		}
	}

} // END class WikiDB_SP_UndefinedTables


//////////////////////////
// Empty Tables

class WikiDB_SP_EmptyTables extends WikiDB_SpecialPage {

	function __construct() {
		parent::pRegister('EmptyTables');
	}

	function GetTextKeyPrefix() {
		return "emptytables";
	}

	function GetQuery(&$DB) {
		$tblTables = $DB->tableName(pWIKIDB_tblTables);
		$tblRows = $DB->tableName(pWIKIDB_tblRows);

		$Fields = array(
					$tblTables . ".table_namespace",
					$tblTables . ".table_title",
				);

		return array(
			'tables' => array(pWIKIDB_tblRows, pWIKIDB_tblTables),
			'fields' => $Fields,
			'conds' => array($tblRows . ".table_title IS NULL"),
			'join_conds' => array(
							pWIKIDB_tblRows => array(
									'LEFT JOIN',
									array(
											$tblRows . ".table_namespace = "
												. $tblTables . ".table_namespace",
											$tblRows . ".table_title = "
												. $tblTables . ".table_title",
										),
								),
							),
			'order' => array(
							$tblTables . ".table_namespace",
							$tblTables . ".table_title",
						),
			'options' => array(
							'GROUP BY'	=> $Fields,
						),
		);
	}

	function pMakeListItem($Row) {
		$Title = Title::makeTitleSafe($Row->table_namespace, $Row->table_title);
		if (!is_null($Title)) {
			$DefLink = $this->pGetLinkToTitle($Title);
			return( "<li>" . $DefLink . '</li>' . "\n");
		}
		else {
			return("<!-- Invalid title " . htmlspecialchars($Row->table_title)
				 . " in namespace " . htmlspecialchars($Row->table_namespace)
				 . " -->\n");
		}
	}

} // END class WikiDB_SP_EmptyTables


//////////////////////////
// All Tables

class WikiDB_SP_AllTables extends WikiDB_SpecialPage {

	function __construct() {
		parent::pRegister('AllTables');
	}

	function GetTextKeyPrefix() {
		return "alltables";
	}

	function GetQuery(&$DB) {
		$tblTables = $DB->tableName(pWIKIDB_tblTables);
		$tblRows = $DB->tableName(pWIKIDB_tblRows);

		$GroupFields = array(
					$tblTables . ".table_namespace",
					$tblTables . ".table_title",
				);

	// Later versions of MW allow us to use 'Alias' => "FieldDef", but to support
	// older versions we need to use "FieldDef AS Alias".
		$Fields = array(
					"COUNT(" . $tblRows . ".table_title) AS record_count",
				);
		$Fields = array_merge($Fields, $GroupFields);

		return array(
			'tables' => array(pWIKIDB_tblRows, pWIKIDB_tblTables),
			'fields' => $Fields,
			'join_conds' => array(
							pWIKIDB_tblRows => array(
									'LEFT JOIN',
									array(
											$tblRows . ".table_namespace = "
												. $tblTables . ".table_namespace",
											$tblRows . ".table_title = "
												. $tblTables . ".table_title",
										),
								),
							),
			'order' => array(
							$tblTables . ".table_namespace",
							$tblTables . ".table_title",
						),
			'options' => array(
							'GROUP BY'	=> $GroupFields,
						),
		);
	}

	function pMakeListItem($Row) {
		$Title = Title::makeTitleSafe($Row->table_namespace, $Row->table_title);
		if (!is_null($Title)) {
			$DefLink = $this->pGetLinkToTitle($Title);
			$DataLink = $Title->getLocalURL('action=viewdata');
			$DataLinkText = WikiDB_Msg('alltables-record', $Row->record_count);

			if ($Row->record_count == 0)
				$Class = ' class="new"';
			else
				$Class = '';

			return("<li>" . $DefLink . ' - <a href="' . $DataLink . '"'
				   . $Class . '>' . $DataLinkText . '</a></li>' . "\n");
		}
		else {
			return("<!-- Invalid title " . htmlspecialchars($Row->table_title)
				 . " in namespace " . htmlspecialchars($Row->table_namespace)
				 . " -->\n");
		}
	}

} // END class WikiDB_SP_AllTables
